__all__ = ['Atom']

class Atom():
    def __init__(self,
                 record_name: str = None,
                 serial: str = None,
                 name: str = None,
                 altLoc: str = None,
                 resName: str = None,
                 chainID: str = None,
                 resSeq: str = None,
                 iCode: str = None,
                 x: str = None,
                 y: str = None,
                 z: str = None,
                 occupancy: str = None,
                 tempFactor: str = None,
                 element: str = None,
                 charge: str = None
        ):
        self.SetRecordName(record_name)
        self.SetSerial(serial)
        self.SetAltLoc(altLoc)
        self.SetResName(resName)
        self.SetChainID(chainID)
        self.SetResSeq(resSeq)
        self.SetiCode(iCode)
        self.SetXYZ(x, y, z)
        self.SetOccupancy(occupancy)
        self.SetTempFactor(tempFactor)
        self.SetElement(element)
        self.SetCharge(charge)
        self.SetName(name) # Do this last to extract the element

    def SetRecordName(self, record_name: str) -> None:
        self.record_name = record_name

    def SetSerial(self, serial: str | int | None) -> None:
        self.serial = int(serial)

    def SetName(self, name) -> None:
        # TODO Extrapolate the element from the name
        self.name = name

    def SetAltLoc(self, altLoc: str | None) -> None:
        self.altLoc = altLoc

    def SetResName(self, resName: str | None) -> None:
        self.resName = resName

    def SetChainID(self, chainID: str | None) -> None:
        self.chainID = chainID

    def SetResSeq(self, resSeq: str | None) -> None:
        self.resSeq = resSeq

    def SetiCode(self, iCode: str | None) -> None:
        self.iCode = iCode

    def SetXYZ(self, x: str | float | None, y: str | float | None, z: str | float | None) -> None:
        self.x = float(x) if x is not None else None
        self.y = float(y) if y is not None else None
        self.z = float(z) if z is not None else None

    def SetX(self, x: str | float | None):
        self.x = float(x) if x is not None else None

    def SetY(self, x: str | float | None):
        self.x = float(x) if x is not None else None

    def SetZ(self, x: str | float | None):
        self.x = float(x) if x is not None else None

    def SetOccupancy(self, occupancy: str | float | None) -> None:
        self.occupancy = float(occupancy) if occupancy is not None else None

    def SetTempFactor(self, tempFactor: str | float | None) -> None:
        self.tempFactor = float(tempFactor) if tempFactor is not None else None

    def SetElement(self, element: str | None) -> None:
        self.element = element

    def SetCharge(self, charge: str | None) -> None:
        self.charge = charge

    def Distance(self, other: 'Atom') -> float:
        return ((self.x - other.x)**2 + (self.y - other.y)**2 + (self.z - other.z)**2)**(.5)

    # In case you want to treat aldepyde like that other importable module that shall not be named
    def __sub__(self, other):
        return self.Distance(other)

    def __str__(self):
        return f"{self.record_name} {self.serial} {self.resName}"





class DataAccessException(Exception):
    def __init__(self, message):
        super().__init__(message)

# Unable to resolve to one output when necessary
class AmbiguousTokenException(Exception):
    def __init__(self, message):
        super().__init__(message)

# Invalid res_type
class BadCategoryException(Exception):
    def __init__(self, message):
        super().__init__(message)

class EntryRetrievalException(Exception):
    def __init__(self, message):
        super().__init__(message)


__version__ = "0.0.0a2"
__author__ = "Nate McMurray"
__description__ = "A module for mangling biomolecules"

from ._config import _configuration
from aldepyde.cache.cachemanager import CacheManager


_cache_manager = CacheManager(initialize=False)
def use_cache(max_memory="2gib", path=None) -> CacheManager:
    global _cache_manager
    _cache_manager = CacheManager(path=path, initialize=True)
    _cache_manager.enable()
    _cache_manager.set_max_memory(max_memory)
    return _cache_manager

def get_cache() -> CacheManager:
    global _cache_manager
    return _cache_manager

from aldepyde.data.store import get_store as gs, del_store as ds, DataStore

def get_store() -> DataStore:
    return gs()

def del_store():
    ds()



from importlib import import_module

__all__ = ["generators", "biomolecule_old", "fetcher"]

def __getattr__(name):
    if name in __all__:
        return import_module(f".{name}", __name__)
    raise AttributeError(name)

def __dir__():
    return sorted(list(globals().keys()) + __all__)


from .polymer_constraints import PolymerConstraints, ConstraintSpecs
from .defaults.protein_defaults import ProteinDefaults

class ProteinConstraints(PolymerConstraints):
    def __init__(self, specs:ConstraintSpecs=None):
        if specs is None:
            default_categories = ProteinDefaults().get_classes()
            cs = ConstraintSpecs(categories=default_categories)
        else:
            cs = specs
        # cs = self.get_default_classes() if classes is None else classes
        super().__init__(specs=cs)


    def get_default_classes(self):
        return ProteinDefaults().get_classes()


import re
import sys
import os
import json
# TODO Consider version information in config
# from aldepyde import __version__

class _configuration():

    class VERSION_SETTINGS():  # There's probably a better way to do this, but it feels neat so I'm keeping it for now
        def __init__(self):
            self.settings = {
                "verbose": False,
                "angle_mode": "degrees",
                "enable_cache": True,
                "cache_path": os.path.join(os.getcwd(), "aldepyde_cache"),
                "cache_max_memory": "500mib"
            }
    def __init__(self):
        s = self.VERSION_SETTINGS()
        self._settings = s.settings
        self._setters = { # TODO Finish fixing this
                "verbose": self.SetVerbose,
                "angle_mode": self.SetAngleMode,
                "enable_cache": self.SetEnableCache,
                "cache_path": self.SetCachePath,
                "cache_max_memory": self.SetCacheMaxMemory
            }

    # def _Verify(self, key, value) -> None:
    #     if value not in self._setters[key][1]:
    #         print(f"Incorrect value for ({key} <-- {value})\n\tAllowed values: {self._setters[key][1]}", file=sys.stderr)
    #         raise ValueError("See above")

    def _Verify(self, key: str, accepted: list, value: object) -> None:
        if True not in [True for x in accepted if value == x or (isinstance(x, type) and isinstance(value, x))]:
            raise ValueError
        # if value not in accepted:
        #     # print(f"Incorrect value for ({key} <-- {value})\n\tAllowed values: {accepted}", file=sys.stderr)
        #     raise ValueError(f"Incorrect value for ({key} <-- {value})\n\tAllowed values: {accepted}")

    def __getitem__(self, item):
        return self._settings[item]

    def GetVerbose(self) -> bool:
        return self._settings['verbose']

    def SetVerbose(self, enabled: bool) -> None:
        # I'm choosing to not be defensive here.
        # Ima be honest, if you break the module with a verbosity boolean, that's on you
        self._settings['verbose'] = enabled

    def GetAngleMode(self) -> str:
        return self._settings['angle_mode']

    def SetAngleMode(self, a_type: str) -> None:
        self._Verify('angle_mode', a_type)
        self._settings['angle_mode'] = a_type

    def SetEnableCache(self, enabled:bool) -> None:
        self._Verify('enable_cache', [bool], enabled)
        self._settings['enable_cache'] = enabled

    def SetCachePath(self, path: str) -> None:
        # TODO Consider making the new cache directory immediately
        self._settings['cache_path'] = path

    def GetCachePath(self) -> str:
        return self._settings['cache_path']

    def GetEnabledCache(self) -> bool:
        return self._settings['enable_cache']

    def SetCacheMaxMemory(self, max_memory: str|int) -> None:
        max_memory = str(max_memory)
        if re.fullmatch("[0-9]*[kmg]i?b", max_memory) is None:
            raise ValueError
        self._settings['cache_max_memory'] = max_memory

    def GetCacheMaxMemory(self) -> str:
        return self._settings['cache_max_memory']

    ####################
    def Load(self, s: dict | str, ignore_missing=False) -> None:
        if isinstance(s, str):
            if os.path.exists(s):  # Try and read as a data file
                with open(s, "r") as jp:
                    s = json.load(jp)
            else:  # Try and read as a data string
                s = json.loads(s)
        extra_settings = "".join([f"\t-{k}\n" for k in s.keys() if k not in self._setters.keys()])
        if len(extra_settings) != 0:
            print(f"Please remove the following config settings:\n{extra_settings}", file=sys.stderr)
            raise KeyError("Input config does not match expectations")
        missing_settings = "".join([f"\t-{k} : {v[1]}\n" for k, v in self._setters.items() if k not in s.keys()])
        if len(missing_settings) != 0 and not ignore_missing:
            print(f"Please include the following config settings:\n{missing_settings}", end="", file=sys.stderr)
            raise KeyError("Input config does not match expectations")
        if ignore_missing:
            vs = self.VERSION_SETTINGS()
            self._settings = vs.settings
        for key in s.keys():
            self._setters[key][0](s[key])

    def GetConfig(self, as_string=False, indent="") -> dict | str:
        if as_string:
            return json.dumps(self._settings, indent=indent)
        return self._settings

    def Save(self, path: str="config.data", indent: str=""):
        with open(path, "w") as jp:
            json.dump(self._settings, jp, indent=indent)




from io import BytesIO
from .pdb_client import pdb_Client
from .parse import parse
import gzip


def fetch_text(code, extension='mmcif') -> str:
    client = pdb_Client()
    raw_gz = client.fetch(code=code, extension=extension)
    with gzip.GzipFile(fileobj=raw_gz) as gz:
        return gz.read().decode('utf-8')

def fetch_raw(code, extension='mmcif') -> BytesIO:
    client = pdb_Client()
    return client.fetch(code=code, extension=extension)

def fetch(code, extension='mmcif'):
    return parse(fetch_text(code=code, extension=extension))


from ._dna import dna
from ._rna import rna
from ._amino_acid import amino_acid
from .Residue import Residue

# These three could probably be condensed
def CheckDNA(self, value: str, *args) -> bool:
    if args == ():
        args = self._dna_map[1]
    map = self.load_values(*args, store_as=None)
    if value in map['dna'].keys():
        return True
    return False


def CheckRNA(self, value: str, *args) -> bool:
    if args == ():
        args = self._map[1]
    map = self.load_values(*args, store_as=None)
    if value in map['dna'].keys():
        return True
    return False


def CheckAA(self, value: str, *args) -> bool:
    if args == ():
        args = self._map[1]
    map = self.load_values(*args, store_as=None)
    if value in map['dna'].keys():
        return True
    return False


def CheckResidue(self, value: str, *args) -> bool:
    if args == ():
        args = self._map[1]
    if self.CheckAA(value, *args):
        return True
    if self.CheckDNA(value, *args):
        return True
    if self.CheckRNA(value, *args):
        return True
    return False


# This method determines if something is DNA, RNA, or an amino acid.
# Don't be cheeky with this. If you aren't following the IUPAC naming schemes,
# you're gonna have a bad time.
#
# RNA has exclusively 1-letter codes: A, C, T, G, etc.
# DNA has exclusively 2-letter codes: DA, DC, DT, DG, etc.
# Amino acids have exclusively 3-letter codes
def ExtrapolateResidueType(self, value: str) -> object:
    if self.CheckRNA(value):
        return rna
    if self.CheckDNA(value):
        return dna
    if self.CheckAA(value):
        return amino_acid
    return Residue